<?php

namespace Drupal\decoupled_pages\Routing;

use Drupal\Component\Assertion\Inspector;
use Drupal\Core\Asset\LibraryDiscoveryInterface;
use Drupal\Core\Routing\RouteBuildEvent;
use Drupal\Core\Routing\RoutingEvents;
use Drupal\decoupled_pages\Exception\RouteDefinitionException;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\Routing\Route;

/**
 * Decorates decoupled page route definitions.
 *
 * @internal
 */
final class RoutingEventSubscriber implements EventSubscriberInterface {

  /**
   * The route default key for indicating a decoupled page route.
   */
  const DECOUPLED_PAGE_ROUTE_DEFAULT_KEY = '_decoupled_page_main';

  /**
   * The route option key for adding assets to a decoupled page route.
   */
  const DECOUPLED_PAGE_ASSETS_ROUTE_OPTION_KEY = '_decoupled_page_assets';

  /**
   * The route default key identifying a list of assets for a decoupled page.
   *
   * This route default is for internal use only.
   */
  const DECOUPLED_PAGE_LIBRARIES_ROUTE_DEFAULT_KEY = 'decoupled_page_libraries';

  /**
   * A library discovery service.
   *
   * @var \Drupal\Core\Asset\LibraryDiscoveryInterface
   */
  protected $libraries;

  /**
   * RoutingEventSubscriber constructor.
   *
   * @param \Drupal\Core\Asset\LibraryDiscoveryInterface $libraries
   *   A library discovery service.
   */
  public function __construct(LibraryDiscoveryInterface $libraries) {
    $this->libraries = $libraries;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[RoutingEvents::ALTER][] = ['alterRoutes'];
    return $events;
  }

  /**
   * Configure `_spa_main` routes.
   *
   * @param \Drupal\Core\Routing\RouteBuildEvent $event
   *   A route build event.
   */
  public function alterRoutes(RouteBuildEvent $event) {
    $route_collection = $event->getRouteCollection();
    foreach ($route_collection as $route_name => $route) {
      if ($route->getDefault(self::DECOUPLED_PAGE_ROUTE_DEFAULT_KEY) !== NULL) {
        $this->alterDecoupledPageRoute($route_name, $route);
      }
    }
  }

  /**
   * Validates and configures a decoupled page route definition.
   *
   * @param string $route_name
   *   The route name.
   * @param Route $route
   *   The route definition object.
   */
  protected function alterDecoupledPageRoute(string $route_name, Route $route) {
    // Validate that the main route library exists.
    $main = $route->getDefault(self::DECOUPLED_PAGE_ROUTE_DEFAULT_KEY);
    $this->checkRouteAsset($route_name, $main);
    // Validate that all route asset libraries exists.
    $assets = $route->getOption(static::DECOUPLED_PAGE_ASSETS_ROUTE_OPTION_KEY);
    if (!(is_array($assets) && Inspector::assertAllStrings($assets))) {
      $format = 'The %s route option, given on the %s route definition, must be an array of strings.';
      throw new RouteDefinitionException(sprintf($format, static::DECOUPLED_PAGE_ASSETS_ROUTE_OPTION_KEY, $route_name));
    }
    foreach ($assets as $asset) {
      $this->checkRouteAsset($route_name, $asset);
    }
    // Combine the main and additional asset libraries from the route definition
    // into a route default to be used by the default controller. First, ensure
    // that the route definition does not already define a value for that route
    //default.
    $assets_route_default = $route->getDefault(static::DECOUPLED_PAGE_LIBRARIES_ROUTE_DEFAULT_KEY);
    if (!is_null($assets_route_default)) {
      $format = 'The %s route default must not be specified on the %s route definition.';
      throw new RouteDefinitionException(sprintf($format, static::DECOUPLED_PAGE_LIBRARIES_ROUTE_DEFAULT_KEY, $route_name));
    }
    $assets_route_argument = [];
    array_push($assets_route_argument, $main, ...$assets);
    $route->setDefault(static::DECOUPLED_PAGE_LIBRARIES_ROUTE_DEFAULT_KEY, $assets_route_argument);
    // If the route does not define a method list, ensure that only `GET` is
    // accepted.
    $methods = $route->getMethods();
    if (empty($methods)) {
      $route->setMethods(['GET']);
    }
    // Decoupled page routes must not define a custom controller.
    if ($route->getDefault('_controller')) {
      $format = 'The %s route definition must not declare a _controller route default when the %s route default is declared.';
      throw new RouteDefinitionException($format, $route_name, static::DECOUPLED_PAGE_ROUTE_DEFAULT_KEY);
    }
    // Set the standard decoupled page controller.
    $route->setDefault('_controller', 'controller.decoupled_pages.main:serve');
  }

  /**
   * Verifies that an asset exists.
   *
   * @param $route_name
   *   The route name requiring the asset.
   * @param $asset
   *   The asset ID.
   */
  protected function checkRouteAsset($route_name, $asset) {
    list($extension, $library_name) = explode('/', $asset, 2);
    $library = $this->libraries->getLibraryByName($extension, $library_name);
    if ($library === FALSE) {
      $format = 'The %s library specified by the %s route does not exist.';
      throw new RouteDefinitionException(sprintf($format, $asset, $route_name));
    }
  }

}
